/*
**	netsock class
**
**	This class prepares a socket to read/write messages
**	and read/write files.
**
**	An error (ERR) returned from any of these I/O methods
**	means that the client has been sent an error message
**	and this socket has been shutdown.
*/

#include <sys/stat.h>
#include <sys/types.h>
#include <windows.h>
#include <winsock.h>
#include <dos.h>

#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <io.h>
#include <stdlib.h>
#include "Tempnam.h"
#include "netsock.h"
#include "network.h"
#include "defines.h"
#include "debuglog.h"

netsock::netsock(int Sock)
{
	Rd      = Sock;
	Wt      = Sock;
	CLOSED  = FALSE;
	MessageSize = 0;
	Message = NULL;
	MsgBuf  = new char[MAXRECSIZE+1];
}

int netsock::NormalMsg(char* msg)
{
	/*
	**	Send severity 0 and message to
	**	the client
	*/
	strcpy(MsgBuf, "0");
	strcat(MsgBuf, DELIMITER);
	strcat(MsgBuf, msg);
	return this->SendMsg(MsgBuf);
}

int netsock::StatusMsg(char* msg)
{
	/*
	**	Send severity 4 and message to
	**	the client
	*/
	strcpy(MsgBuf, "4");
	strcat(MsgBuf, DELIMITER);
	strcat(MsgBuf, msg);
	return this->SendMsg(MsgBuf);
}

int netsock::WarningMsg(char* msg)
{
	/*
	**	Send severity 3 and message to
	**	the client
	*/
	strcpy(MsgBuf, "3");
	strcat(MsgBuf, DELIMITER);
	strcat(MsgBuf, msg);
	return this->SendMsg(MsgBuf);
}

void netsock::Abort(char* errmsg)
{
	/*
	**	Send severity 1 and message to
	**	the client then close socket.
	*/
	stripLine(errmsg);
	strcpy(MsgBuf, "1");
	strcat(MsgBuf, DELIMITER);
	strcat(MsgBuf, errmsg);
	this->SendMsg(MsgBuf);
	Sleep(4000);
	//debuglog("Abort: message (%s) sent to client.\n", MsgBuf);
	this->Close();
}

void netsock::Close(void)
{
	if (!CLOSED)
	{
		close(Rd);
		CLOSED = TRUE;
	}
}

int netsock::SendMsg(char* msg)
{
	char	buf[10];

	if(!msg)
	{
		return (ERR);
	}

	/*
	**	Get the size of the message.  Store a newline at the END
	**	of the size buffer since our protocol is to read 10 full
	**	characters for the size of the message.
	*/

	int size = strlen(msg) + 1;
	strcpy(buf, "         ");
	sprintf(buf, "%d", size);
	buf[9] = '\n';

	/*
	**	Send the length of msg in first 10 bytes.
	*/
	int ret = this->Send(buf, 10);
	if(ret == ERR)
	{
		return (ERR);
	}

	if(size-1 > 0)
	{
		ret = this->Send(msg, size-1);
		if (ret == ERR)
		{
			return (ERR);
		}
	}
	ret = this->Send("\n", 1);
	if (ret == ERR)
	{
		return (ERR);
	}
	return (OK);
}

int netsock::RecvMsg(void)
{
	char buf[10];
	
	/*
	**	Get the size of the message from the 1st 
	**	10 bytes.
	*/
	int ret = this->Recv(buf, 10);
	if (ret == ERR)
	{
		debuglog("RecvMsg: Recv() message size failed.\n");
		return (ERR);
	}
	if (buf[9] != '\n')
	{
		debuglog("RecvMsg: no newline reading size of message.\n");
		this->Abort("No newline protocol on received network msg.");
		return (ERR);
	}

	buf[9] = '\0';	// convert raw message to string.

	/*
	**	Make sure the buffer to read the message into
	**	is large enough.
	*/
	int size = atoi(buf);
	if(size <= 0)
	{
		debuglog("RecvMsg: error getting message size.\n");
		this->Abort("Network error getting message size.");
		return (ERR);
	}

	if (size > MessageSize)
	{
		if (Message)
		{
			delete Message;
		}
		Message = new char[size+1];
		MessageSize = size;
	}
	firstToken = TRUE;


	ret = this->Recv(Message, size);
	if (ret == ERR)
	{
		debuglog("RecvMsg: Recv() message body (%d) failed.\n",
			size);
		return (ERR);
	}

	if (Message[size-1] != '\n')
	{
		debuglog("RecvMsg: no newline at end of message.\n");
		this->Abort("No newline protocol on received network msg.");
		return (ERR);
	}
	Message[size-1] = '\0';
	return(OK);
}

char* netsock::getMsg(void)
{
	return (Message);
}

char* netsock::getToken(void)
{
	if (firstToken)
	{
		firstToken = FALSE;
		return (strtoken(Message, DELIMITER));
	}
	else
	{
		return (strtoken(NULL, DELIMITER));
	}
}

int netsock::SendFile(char* file)
{
	int	ret;
	int	amt;
	int	bsize = 0;		// bytes to send
	int	fsize = 0;
	struct	stat	ufstat;
	int	totbytes = 0;
	int	rmode = O_RDONLY;

	umask(0022);

	/*
	**	Get the size of the file to send.
	*/
	if(stat(file, &ufstat) != 0)
	{
		debuglog("SendFile: stat failed on \"%s\", err=%d.\n",
			file, errno);
		this->Abort("Can't get file size.");
		return (ERR);
	}
	fsize = (int) ufstat.st_size;

	/*
	**	On UNIX, file size is always total bytes to send.
	*/
	bsize = fsize;
	rmode = O_RDONLY|O_BINARY;

	/*
	**	First, send the protocol RETURN_FILE
	*/
	ret = this->SendMsg(RETURN_FILE);
	if (ret == ERR) 
	{
		debuglog("SendFile: SendMsg(RETURN_FILE) failed.\n");
		return (ret);
	}

	/*
	**	Next, send the total bytes to be transmitted.
	*/
	sprintf(MsgBuf, "%d\n", bsize);
	ret = this->SendMsg(MsgBuf);
	if (ret == ERR) 
	{
		return (ret);
	}

	/*
	**	Open the file for reading.
	*/
	int f;
	if ((f = open(file, rmode)) < 0)
	{
		debuglog("SendFile: can't open file %s\n", file);
		this->Abort("Can't open file to send.");
		return (ERR);
	}

	/*
	**	Now send the bytes in the file.
	*/
	for (int i = 0; i < fsize; i += MAXRECSIZE)
	{
		amt = MAXRECSIZE;
		if ((i + amt) > fsize) amt = fsize - i;

		ret = read(f, MsgBuf, amt);
		if(ret <= 0)
		{
			debuglog("SendFile: file read failed, errno=%d\n",
				errno);
			close(f);
			this->Abort("Can't read file trying to send.");
			return (ERR);
		}

		MsgBuf[ret] = '\0';
		totbytes += ret;
		int r = this->Send(MsgBuf, ret);
		if(r == ERR)
		{
			close(f);
			debuglog("SendFile: Send() failed, errno=%d\n", r);
			return (ERR);
		}
	}

	(void) close(f);

	/*
	**	Sanity check on the read operation.
	*/
	if (totbytes != bsize)
	{
		debuglog("SendFile: file changed size:\n");
		debuglog(
		     "SendFile: supposed to send %d but sent %d (fsize=%d).\n",
		     bsize, totbytes, fsize);
		this->Abort("Incorrect file size sent.");
		return (ERR);
	}

	/*
	**	Last, wait for protocol ACK
	*/
	ret = this->RecvMsg();
	if (ret == ERR) 
	{
		debuglog("SendFile: RecvMsg(ACK) failed.\n");
		return (ret);
	}
	if(strcmp(Message, CACK_MSG))
	{
		debuglog("SendFile: didn't get ACK, got %s.\n", Message);
		this->Abort("Incorrect SendFile protocol, expected ACK.");
		return (ERR);
	}

	return(OK);
}

int netsock::PassFile(netsock* Client)
{
	int ret = this->RecvMsg();
	if (ret == ERR) 
	{
		debuglog("PassFile: RecvMsg(RETURN_FILE) failed.\n");
		return (ret);
	}
	if(!strcmp(Message, NO_RETURN_FILE))
	{
		ret = Client->SendMsg(NO_RETURN_FILE);
		if (ret == ERR)
		{
			debuglog(
			"PassFile: Client SendMsg(NO_RETURN_FILE) failed.\n");
			return (ERR);
		}
		return (ERR);
	}

	ret = Client->SendMsg(RETURN_FILE);
	if (ret == ERR)
	{
		debuglog("PassFile: Client SendMsg(RETURN_FILE) failed.\n");
		return (ERR);
	}

	/*
	**	Now, get the size of the file to receive from
	**	network header message.
	*/
	ret = this->RecvMsg();
	if (ret == ERR)
	{
		debuglog("PassFile: can't get file size.\n");
		return (ERR);
	}

	int size = atoi(Message);
	if (size < 0)	//  could get zero-length file
	{
		debuglog("PassFile: received wrong msg on socket.\n");
		this->Abort("Received wrong msg on socket.");
		Client->Abort("Received wrong msg on socket.");
		return (ERR);
	}

	ret = Client->SendMsg(Message);
	if (ret == ERR)
	{
		debuglog("PassFile: Client SendMsg() failed.\n");
		return (ERR);
	}

	/*
	**	Receive file data into our local file.
	*/
	for (int i = 0; i < size; i += MAXRECSIZE)
	{
		/*
		**	Read a chunk at a time from network
		**	adjusting the size of the last chunk
		**	to finish off read.  Copy to file on
		**	the server.
		*/
		int amt = MAXRECSIZE;
		if(i + amt > size) amt = size - i;

		ret = this->Recv(MsgBuf, amt);
		if (ret == ERR)
		{
			debuglog("PassFile: file error, errno=(%d)!\n",
				 errno);
			Client->Abort("Error reading net data.");
			return (ERR);
		}
		MsgBuf[amt] = '\0';

		/*
		**	Dump the buffer read to the Client.
		*/
		if (ret > 0)
		{
			if(Client->Send(MsgBuf, ret) == ERR)
			{
				debuglog("PassFile: Client Send(%d) failed\n",
					ret);
				Client->Abort("Error sending net data.");
				return (ERR);
			}
		}
	}

	/*
	**	Last, wait for protocol ACK from Client
	*/
	ret = Client->RecvMsg();
	if (ret == ERR) 
	{
		debuglog("PassFile: RecvMsg(ACK) failed.\n");
		return (ret);
	}
	if(strcmp(Client->getMsg(), CACK_MSG))
	{
		debuglog("PassFile: didn't get ACK, got %s.\n", Message);
		this->Abort("Incorrect PassFile protocol, expected ACK.");
		Client->Abort("Protocol error getting file from base server.");
		return (ERR);
	}

	ret = this->SendMsg(CACK_MSG);
	if (ret == ERR)
	{
		debuglog("PassFile: server SendMsg() failed.\n");
		return (ERR);
	}

	return (OK);
}

int netsock::RecvFile(char* file)
{
	int ret = this->RecvMsg();
	if (ret == ERR) 
	{
		debuglog("RecvFile: RecvMsg(RETURN_FILE) failed.\n");
		return (ret);
	}
	if(!strcmp(Message, NO_RETURN_FILE))
	{
		return (ERR);
	}
	else
	if(strcmp(Message, RETURN_FILE))
	{
		debuglog("RecvFile: didn't get RETURN_FILE, got %s.\n",
			Message);
		return (ERR);
	}
	ret = this->RecvFileNP(file);
	if (ret == ERR) 
	{
		debuglog("RecvFile: RecvFileNP failed.\n");
		return (ret);
	}

	/*
	**	Last, send the protocol ACK
	*/
	ret = this->SendMsg(CACK_MSG);
	if (ret == ERR) 
	{
		debuglog("RecvFile: SendMsg(CACK_MSG) failed.\n");
		return (ret);
	}

	return (OK);
}

int netsock::RecvFileNP(char* file)	// No Protocol
{
	/*
	**	Reads the socket writing all data to the file
	**	until the entire file is read.  The amount of
	**	data to read is determined by the first message
	**	which contains the file size.
	**
	*/
	umask(0022);

	/*
	**	Open file to receive network data for writing.
	*/
	int f;
	int rmode = O_WRONLY|O_CREAT|O_BINARY;
	if((f = open(file, rmode, 0644)) < 0)
	{
		debuglog("RecvFileNP: can't open file=%s, errno=%d\n",
			file, errno);
		this->Abort("Can't open file to receive network data.");
		return (ERR);
	}

	/*
	**	Now, get the size of the file to receive from
	**	network header message.
	*/
	int ret = this->RecvMsg();
	if (ret == ERR)
	{
		debuglog("RecvFileNP: can't get file size.\n");
		close(f);
		unlink(file);
		return (ERR);
	}

	int size = atoi(Message);
	if (size < 0)	//  could get zero-length file
	{
		debuglog("RecvFileNP: received wrong msg on socket.\n");
		close(f);
		unlink(file);
		this->Abort("Received wrong msg on socket.");
		return (ERR);
	}

	/*
	**	Receive file data into our local file.
	for (int i = 0; i < size; i += MAXRECSIZE)
	*/
	int amt = 0;
	while (amt < size)
	{
		/*
		**	Read a chunk at a time from network
		**	adjusting the size of the last chunk
		**	to finish off read.  Copy to file on
		**	the server.
		*/
		int tryfor = MAXRECSIZE;
		if(tryfor + amt > size) tryfor = size - amt;

		ret = this->Recv(MsgBuf, tryfor);
		if (ret == ERR)
		{
			debuglog("RecvFileNP: Recv failed, ret=(%d)!\n",ret);
			close(f);
			unlink(file);
			return (ERR);
		}
		MsgBuf[ret] = '\0';

		/*
		**	Dump the buffer read to the file.
		*/
		if (ret > 0)
		{
			if (write(f, MsgBuf, ret) != ret)
			{
			    debuglog("RecvFileNP: file error, errno=(%d)!\n",
					 errno);
				close(f);
				unlink(file);
				this->Abort("Error writing net data to file.");
				return (ERR);
			}
		}
		amt += ret;
	}

	close(f);
	return (OK);
}

int netsock::Recv(char* msg, int size)
{
	/*
	**	network socket read
	*/
	int ret;
	int done = FALSE;
	fd_set	fset;
	struct timeval tval;
	tval.tv_sec = 0;
	tval.tv_usec = 0;

	while (!done)
	{
		FD_ZERO(&fset);
		FD_SET(Rd, &fset);
		ret = select(1, &fset, NULL, NULL, &tval);
		if(ret > 0)
		{
			ret = recv(Rd, msg, size, 0);
			done = TRUE;
		}
		else if(ret == SOCKET_ERROR)
		{
			debuglog( "Recv: select err ret=%d, err=%d\n",
				ret, WSAGetLastError());
			done = TRUE;
		}
	}
	return (ret);
}

int netsock::Send(char* msg, int size)
{
	int ret;
	int done = FALSE;
	fd_set	fset;
	struct timeval tval;

	while (!done)
	{
		FD_ZERO(&fset);
		FD_SET(Wt, &fset);
		tval.tv_sec = 0;
		tval.tv_usec = 0;
		ret = select(1, NULL, &fset, NULL, &tval);
		if(ret > 0)
		{
			ret = send(Wt, msg, size, 0);
			if(ret == SOCKET_ERROR)
			{
				int Err = WSAGetLastError();
				if (Err == WSAEWOULDBLOCK)
				{
					done = FALSE;
				}
				else
				{
					debuglog(
					   "Send: send failed err=%d\n", Err);
					done = TRUE;
				}
			}
			else
			{
				done = TRUE;
			}
		}
		else if(ret == SOCKET_ERROR)
		{
			if (WSAGetLastError() == WSAEINPROGRESS)
			{
				debuglog(
				    "Send:select: xfer in progress.\n");
				done = FALSE;
			}
			else
			{
				debuglog(
					"Send: select err ret=%d, err=%d\n",
					ret, WSAGetLastError());
				done = TRUE;
			}
		}
		else
		{
			done = FALSE;
		}
	}
	return ret;
}

netsock::~netsock()
{
	this->Close();
	delete MsgBuf;

	if (Message)
	{
		delete Message;
	}
}

char* stripMsg(char* msg)
{
	char*	p = strchr(msg, DELIMITER_CHAR);
	if (p)
	{
		return p+1;
	}
	else
	{
		return msg;
	}
}

char* stripLine(char* msg)
{
	char *p = msg;

	for (; *p != '\0'; p++)
	{
		if (*p == '\r' || *p == '\n')
		{
			*p = ' ';
		}
	}
	return msg;
}

